""" Entry point for Next-action's command-line interface. """

import fileinput

from next_action.arguments import parse_arguments
from next_action.pick_action import next_actions
from next_action.todotxt import read_todotxt_file


def next_action() -> None:
    """ Entry point for the command-line interface.

        Basic recipe:
        1) parse command-line arguments,
        2) read todo.txt file(s),
        3) determine the next action(s) and display them.
    """
    arguments = parse_arguments()
    try:
        with fileinput.input(arguments.filenames) as todotxt_file:
            tasks = read_todotxt_file(todotxt_file)
    except OSError as reason:
        arguments.parser.error("can't open file: {0}".format(reason))
    actions = next_actions(tasks, *arguments.filters)
    print("\n".join(action.text for action in actions[:arguments.number]) if actions else "Nothing to do!")



""" Package for parsing command line arguments. """

import argparse
import os
import shutil
import sys
from typing import cast, Dict, List, Set, Tuple

from .arguments import Arguments
from .parser import build_parser, parse_remaining_args


def parse_arguments() -> Arguments:
    """ Build the argument parser, paerse the command line arguments, and post-process them. """
    # Ensure that the help info is printed using all columns available
    os.environ['COLUMNS'] = str(shutil.get_terminal_size().columns)
    default_filenames = ["~/todo.txt"]
    parser = build_parser(default_filenames)
    arguments = Arguments(parser, default_filenames)
    namespace, remaining = parser.parse_known_args()
    parse_remaining_args(parser, remaining, namespace)
    arguments.filenames = namespace.filenames
    arguments.number = namespace.number
    arguments.show_all(namespace.all)
    arguments.filters = namespace.filters
    return arguments



""" Parser for the command line arguments. """

import argparse
from typing import List

import next_action


def build_parser(default_filenames: List[str]) -> argparse.ArgumentParser:
    """ Create the arguments parsers. """
    parser = argparse.ArgumentParser(
        description="Show the next action in your todo.txt. The next action is selected from the tasks in the todo.txt \
                     file based on priority, due date, creation date, and supplied filters.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        usage="%(prog)s [-h] [--version] [-f <filename>] [-n <number> | -a] [<context|project> ...]")
    add_optional_arguments(parser, default_filenames)
    add_positional_arguments(parser)
    return parser


def add_optional_arguments(parser: argparse.ArgumentParser, default_filenames: List[str]) -> None:
    """ Add the optional arguments to the parser. """
    parser.add_argument(
        "--version", action="version", version="%(prog)s {0}".format(next_action.__version__))
    parser.add_argument(
        "-f", "--file", action="append", dest="filenames", metavar="<filename>", default=default_filenames, type=str,
        help="filename of todo.txt file to read; can be - to read from standard input; argument can be repeated to "
             "read tasks from multiple todo.txt files")
    number = parser.add_mutually_exclusive_group()
    number.add_argument(
        "-n", "--number", metavar="<number>", help="number of next actions to show", type=int, default=1)
    number.add_argument("-a", "--all", help="show all next actions", action="store_true")


def add_positional_arguments(parser: argparse.ArgumentParser) -> None:
    """ Add the positional arguments to the parser. """
    filters = parser.add_argument_group("optional context and project arguments; these can be repeated")
    # Collect all context and project arguments in one list:
    filters.add_argument("filters", metavar="<context|project>", help=argparse.SUPPRESS, nargs="*", type=filter_type)
    # Add two dummy arguments for the help info:
    filters.add_argument("dummy", metavar="@<context>", help="context the next action must have",
                         nargs="*", default=argparse.SUPPRESS)
    filters.add_argument("dummy", metavar="+<project>", help="project the next action must be part of",
                         nargs="*", default=argparse.SUPPRESS)
    # These arguments won't be collected because they start with a -. They'll be parsed by parse_remaining_args below
    filters.add_argument("dummy", metavar="-@<context>", help="context the next action must not have",
                         nargs="*", default=argparse.SUPPRESS)
    filters.add_argument("dummy", metavar="-+<project>", help="project the next action must not be part of",
                         nargs="*", default=argparse.SUPPRESS)


def parse_remaining_args(parser: argparse.ArgumentParser, remaining: List[str], namespace: argparse.Namespace) -> None:
    """ Parse the remaining command line arguments. """
    for value in remaining:
        if value.startswith("-@") or value.startswith("-+"):
            argument_type = "context" if value.startswith("-@") else "project"
            if len(value) == 2:
                parser.error("argument <context|project>: {0} name missing".format(argument_type))
            elif value[1:] in namespace.filters:
                parser.error("{0} {1} is both included and excluded".format(argument_type, value[2:]))
            else:
                namespace.filters.append(value)
        else:
            parser.error("unrecognized arguments: {0}".format(value))


def filter_type(value: str) -> str:
    """ Return the filter if it's valid, else raise an error. """
    if value.startswith("@") or value.startswith("+"):
        if len(value) > 1:
            return value
        else:
            value_type = "context" if value.startswith("@") else "project"
            raise argparse.ArgumentTypeError("{0} name missing".format(value_type))
    raise argparse.ArgumentTypeError("unrecognized arguments: {0}".format(value))



""" Algorithm for deciding the next action(s). """

import datetime
from typing import Set, Sequence, Tuple

from .todotxt import Task


def sort_key(task: Task) -> Tuple[str, datetime.date, datetime.date]:
    """ Return the sort key for a task. """
    return (task.priority() or "ZZZ", task.due_date() or datetime.date.max, task.creation_date() or datetime.date.max)


def next_actions(tasks: Sequence[Task], contexts: Set[str] = None, projects: Set[str] = None,
                 excluded_contexts: Set[str] = None, excluded_projects: Set[str] = None) -> Sequence[Task]:
    """ Return the next action(s) from the collection of tasks. """
    # First, get the potential next actions by filtering out completed tasks and tasks with a future creation date
    actionable_tasks = [task for task in tasks if task.is_actionable()]
    # Then, exclude tasks that have an excluded context
    eligible_tasks = filter(lambda task: not excluded_contexts & task.contexts() if excluded_contexts else True,
                            actionable_tasks)
    # And, tasks that have an excluded project
    eligible_tasks = filter(lambda task: not excluded_projects & task.projects() if excluded_projects else True,
                            eligible_tasks)
    # Then, select the tasks that belong to all given contexts, if any
    tasks_in_context = filter(lambda task: contexts <= task.contexts() if contexts else True, eligible_tasks)
    # Next, select the tasks that belong to at least one of the given projects, if any
    tasks_in_project = filter(lambda task: projects & task.projects() if projects else True, tasks_in_context)
    # Finally, sort by priority, due date and creation date
    return sorted(tasks_in_project, key=sort_key)



""" Argument data class for transfering command line arguments. """

import argparse
import os.path
import sys
from typing import List, Set


class Arguments(object):
    """ Argument data class. """
    def __init__(self, parser: argparse.ArgumentParser, default_filenames: List[str]) -> None:
        self.parser = parser
        self.__default_filenames = default_filenames
        self.__filenames: List[str] = []
        self.__filters: List[Set[str]] = []
        self.number = 1

    @property
    def filenames(self) -> List[str]:
        """ Return the filenames. """
        return self.__filenames

    @filenames.setter
    def filenames(self, filenames: List[str]) -> None:
        """ Set the filenames. """
        # Work around the issue that the "append" action doesn't overwrite defaults.
        # See https://bugs.python.org/issue16399.
        if self.__default_filenames != filenames:
            for default_filename in self.__default_filenames:
                filenames.remove(default_filename)
        # Remove duplicate filenames while maintaining order.
        self.__filenames = [os.path.expanduser(filename) for filename in list(dict.fromkeys(filenames))]

    def show_all(self, show_all: bool) -> None:
        """ If the user wants to see all next actions, set the number to something big. """
        if show_all:
            self.number = sys.maxsize

    @property
    def filters(self) -> List[Set[str]]:
        """ Return the filters: contexts, projects, excluded_contexts, excluded_projects. """
        return self.__filters

    @filters.setter
    def filters(self, filters: List[str]) -> None:
        """ Process the context and projects into separate filters. """

        def subset(filters: List[str], prefix: str) -> Set[str]:
            """ Return a subset of the filters based on prefix. """
            return set([f.strip(prefix) for f in filters if f.startswith(prefix)])

        self.__filters = [subset(filters, prefix) for prefix in ("@", "+", "-@", "-+")]
